module Horn (encodeTPTP, encodeINF, encodeINF_noCondition, splitIf, tupling, trueSymbol, falseSymbol) where

import Data.List
import Text.Read
import Terms
import Rules
import TPTP
import ParserForINF

-- Checking if a Horn cluase admits a trivial model.

-- Satisfiability checker of propositional Horn formulas

deleteAll x ys = [ y | y <- ys, y /= x ]

satisfiable :: Eq a => [HornClause a] -> Bool
satisfiable cs = satisfiable1 cs cs

satisfiable1 cs1 []                   = True
satisfiable1 cs1 (([], Nothing) : cs) = False
satisfiable1 cs1 (([], y@(Just a)) : cs) = satisfiable cs2 
  where cs2 = [ (deleteAll a xs, z) | (xs, z) <- cs1, y /= z ]
satisfiable1 cs1 (_ : cs) = satisfiable1 cs1 cs

encodeTPTP :: TPTP -> Maybe [HornClause Atom]
encodeTPTP tptp = sequenceA [ encodeClause clause | CNF _ _ clause <- tptp ]

encodeINF :: INFProblem -> [HornClause Atom]
encodeINF (SemiEquational ceses)   = infToHorn ceses
encodeINF (Join ceses)             = infToHorn ceses
encodeINF (Oriented ceses)         = infToHorn ceses
infToHorn :: (CES, ES) -> [HornClause Atom]
infToHorn (ces, es) = [ (toAtoms es', Just (Equation l r)) | (es', (l, r)) <- ces ++ [ceqForCondition es] ] 
                   ++ [([Equation trueTerm falseTerm], Nothing)]
  where ceqForCondition es = (es, (trueTerm, falseTerm))
        toAtoms es = [ Equation l r | (l, r) <- es ]

encodeINF_noCondition :: INFProblem -> [HornClause Atom]
encodeINF_noCondition (SemiEquational ceses)   = infToHorn_noCondition ceses
encodeINF_noCondition (Join ceses)             = infToHorn_noCondition ceses
encodeINF_noCondition (Oriented ceses)         = infToHorn_noCondition ceses
infToHorn_noCondition :: (CES, ES) -> [HornClause Atom]
infToHorn_noCondition (ces, es) = infToHorn (deleteConditions ces, es)
  where deleteConditions ces = [ ([], e0) | (es0, e0) <- ces ] 

        
encodeClause []                               = Just ([], Nothing)
encodeClause ((True, z) : clause)
  | Just (xs, Nothing) <- encodeClause clause = Just (xs, Just z)
encodeClause ((False, z) : clause)
  | Just (xs, ys) <- encodeClause clause      = Just (z : xs, ys)
encodeClause _ = Nothing 


rootOfPredicate (Predicate (F f _)) = Just f
rootOfPredicate (Equation _ _)      = Nothing

propositionalEncoding hornClauses =
  [ ([ f | Predicate (F f _) <- ts ], Nothing)
  | (ts, Nothing) <- hornClauses ] ++
  [ ([ f | Predicate (F f _) <- ts ], Just g)
  | (ts, Just (Predicate (F g _))) <- hornClauses ]

hasTrivialModel :: [HornClause Atom] -> Bool
hasTrivialModel hornClauses = satisfiable (propositionalEncoding hornClauses)

-- Trasformation from Horn clauses to word problems
-- (Claessen and Smallbone, IJCAR 2018)

trueSymbol  = "true__"
falseSymbol = "false__"
trueTerm    = F trueSymbol []
falseTerm   = F falseSymbol []

liftAtom :: Atom -> (Term, Term)
liftAtom (Predicate t)  = (t, trueTerm)
liftAtom (Equation s t) = (s, t)

liftHornClause :: HornClause Atom -> CEquation
liftHornClause (ps, Nothing) =
  ([ liftAtom p | p <- ps], (falseTerm, trueTerm))
liftHornClause (ps, Just e)  =
  ([ liftAtom p | p <- ps], liftAtom e)

liftHornClauses :: [HornClause Atom] -> CES
liftHornClauses hornClauses = [ liftHornClause c | c <- hornClauses ]

fresh :: Int -> String
fresh n = "f" ++ show n

functionsInCES ces = functionsInES [ e' | (es, e) <- ces, e' <- e : es ]

fSuffix :: String -> Maybe Int
fSuffix ('f' : s) = readMaybe s
fSuffix _         = Nothing

fIndex ces =
  foldl max 0 [ n | f <- functionsInCES ces, Just n <- [fSuffix f] ] + 1


tuplingCES' :: Int -> CES -> ES
tuplingCES' n [] = []
tuplingCES' n (([], e) : hornClauses) = e : tuplingCES' n hornClauses
tuplingCES' n ((es, (u, v)) : hornClauses) = 
  (F f (esL ++ ws), u) : (F f (esR ++ ws), v) : tuplingCES' (n + 1) hornClauses
  where
    f = fresh n
    esL = [ l | (l, _) <- es ]
    esR = [ r | (_, r) <- es ]
    ws = [ V x | x <- nub (vs ++ vars u ++ vars v) ]
    vL = [ v | t <- esL, v <- vars t]
    vR = [ v | t <- esR, v <- vars t]
    vs = if length vL < length vR then vL else vR
    
tuplingCES :: CES -> ES
tuplingCES hornClauses = tuplingCES' (fIndex hornClauses) hornClauses

tupling :: [HornClause Atom] -> (ES, Equation)
tupling hornClauses
  = if hasTrivialModel hornClauses then
      ([], (trueTerm, falseTerm))
    else
      (tuplingCES (liftHornClauses hornClauses), (trueTerm, falseTerm))


splitIfCES' :: Int -> CES -> ES
splitIfCES' n [] = []
splitIfCES' n (([], e) : hornClauses) = e : splitIfCES' n hornClauses
splitIfCES' n (((s0, t0) : es, (u, v)) : hornClauses) =
  (F f (s : ws), u) : splitIfCES' (n + 1) ((es, (F f (t : ws), v)) : hornClauses)
  where
    f = fresh n
    ws = [ V x | x <- nub (vars s ++ vars u ++ vars v) ]
    (s, t) = if length (vars s0) < length (vars t0) then (s0, t0) else (t0, s0)

splitIfCES :: CES -> ES
splitIfCES hornClauses = splitIfCES' (fIndex hornClauses) hornClauses

splitIf :: [HornClause Atom] -> (ES, Equation)
splitIf hornClauses
  = if hasTrivialModel hornClauses then
      ([], (trueTerm, falseTerm))
    else
      (splitIfCES (liftHornClauses hornClauses), (trueTerm, falseTerm))
